SystemOrganization addCategory: #'OLPC-Display'!

DisplayScreen subclass: #OLPCVirtualScreen
	instanceVariableNames: 'display warpBlt pendingUpdate zoomOut'
	classVariableNames: 'VirtualScreenExtent'
	poolDictionaries: ''
	category: 'OLPC-Display'!

!OLPCVirtualScreen commentStamp: '<historical>' prior: 0!
A custom display screen to allow displaying the high-res OLPC content on a regular display.!

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 9/5/2008 14:43'!
defaultVirtualScreenExtent

	^ 1200@900.
! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 5/9/2007 12:23'!
initialize
	"OLPCVirtualScreen initialize"
	Preferences addPreference: #enableVirtualOLPCDisplay categories: #('OLPC') default: false balloonHelp: 'If true, uses the virtual OLPC resolution (1200x900)' projectLocal: true changeInformee: self changeSelector: #virtualDisplayChanged.
	Preferences addPreference: #enablePortraitMode categories: #('OLPC') default: false balloonHelp: 'If true, portrait mode instead of landscape' projectLocal: true changeInformee: self changeSelector: #portraitModeChanged.
	"Smalltalk addToStartUpList: self."
	"self startUp."
	self virtualScreenExtent: 1200@900.
! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'ar 4/12/2007 16:59'!
install
	"Install me and my sensor"
	Display := self new.
	Display beDisplay.
	OLPCVirtualSensor install.! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'ar 4/18/2007 10:16'!
portraitModeChanged
	"Toggle portrait mode"
	(Display isKindOf: self) ifTrue:[Display initialize].! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 8/6/2008 10:02'!
preferredScreenDepth

	^ 16.
! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 11/20/2009 11:11'!
setupIfNeeded

	self actualScreenSize = self virtualScreenExtent
		ifFalse: [
			(Display class ~= self or: [Display extent ~= self virtualScreenExtent]) ifTrue: [
				self install]]
		ifTrue: [Display class = self ifTrue: [self unInstall]]! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 5/9/2007 10:20'!
startUp
	"Figure out if we run on OLPC or not and need a virtual display"
	"Preferences enableVirtualOLPCDisplay ifTrue:[self install]."! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 8/6/2008 10:06'!
unInstall
	"Uninstall me and my sensor"
	EventSensor install.
	Display := DisplayScreen extent: DisplayScreen actualScreenSize depth: self preferredScreenDepth.
	Display beDisplay.! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'ar 4/12/2007 21:22'!
virtualDisplayChanged
	"Update the display"
	Preferences enableVirtualOLPCDisplay 
		ifTrue:[self install]
		ifFalse:[self unInstall].! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 9/5/2008 14:43'!
virtualScreenExtent

	^ VirtualScreenExtent ifNil: [self defaultVirtualScreenExtent].
! !

!OLPCVirtualScreen class methodsFor: 'installing' stamp: 'yo 5/9/2007 10:39'!
virtualScreenExtent: aPoint

	VirtualScreenExtent _ aPoint
! !

!OLPCVirtualScreen methodsFor: 'private' stamp: 'bf 5/8/2010 11:49'!
actualDisplay
	"The display known to the VM"
	^display ifNil: [self]! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'ar 4/13/2007 09:53'!
beDisplay
	"Make the receiver be the main display"
	display ifNil:[super beDisplay] ifNotNil:[display beDisplay].! !

!OLPCVirtualScreen methodsFor: 'copying' stamp: 'ar 4/17/2007 14:33'!
border: rect width: borderWidth rule: rule fillColor: fillColor
	super border: rect width: borderWidth rule: rule fillColor: fillColor.
	self forceToScreen: rect.! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'yo 5/9/2007 11:52'!
canZoomOut

	^ display extent x > self extent x and: [display extent y > self extent y]
! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'yo 5/9/2007 11:37'!
checkForNewScreenSize
	| aPoint |
	aPoint := DisplayScreen actualScreenSize.
	aPoint = display extent ifTrue:[^nil].
	display setExtent: aPoint depth: depth.
	display fillColor: (Color gray: 0.2). 
	self setupWarp; forceToScreen.
	display forceToScreen. "to capture the borders"
	World restoreMorphicDisplay.
	World repositionFlapsAfterScreenSizeChange.! !

!OLPCVirtualScreen methodsFor: 'copying' stamp: 'ar 4/12/2007 21:49'!
copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: aForm
	"We need this for the project zoom"
	super copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: aForm.
	self forceToScreen: (destOrigin extent: sourceRect extent).! !

!OLPCVirtualScreen methodsFor: 'copying' stamp: 'yo 8/29/2007 18:20'!
copy: aRect
 	"Return a new form which derives from the portion of the original form delineated by aRect."
	| newForm |
	newForm _ DisplayScreen extent: aRect extent depth: depth.
	^ newForm copyBits: aRect from: self at: 0@0
		clippingBox: newForm boundingBox rule: Form over fillColor: nil! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'ar 4/17/2007 14:19'!
displayOffset
	"Answer the offset to apply to events for this display"
	display ifNil:[^0@0].
	^(display extent - (self extent / self displayScale)) // 2
! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'yo 5/9/2007 11:59'!
displayScale
	"Answer the scale to apply to events for this display"
	display ifNil:[^1].
	(zoomOut and: [self canZoomOut]) ifTrue: [^ 1].

	^ width / display width max: height / display height! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'ar 4/12/2007 21:15'!
forceDisplayUpdate
	display ifNil:[super forceDisplayUpdate] ifNotNil:[display forceDisplayUpdate].! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'ar 4/20/2007 11:05'!
forceToScreen: rect
	"Update the display"
	| dstRect scale |
	display ifNil:[^super forceToScreen: rect].
	scale := self displayScale.
	"Note: We specifically overestimate the clip rect to avoid roundoff problems"
	dstRect := rect intersect: self boundingBox.
	dstRect := (dstRect origin // scale) truncated - 1 corner: (dstRect corner // scale) truncated + 1.
	dstRect := dstRect translateBy: self displayOffset.
	warpBlt clipRect: dstRect.
	warpBlt warpBits.
	display forceToScreen: dstRect.
	pendingUpdate ifTrue:[
		pendingUpdate := false.
		display forceToScreen.
	].! !

!OLPCVirtualScreen methodsFor: 'initialize' stamp: 'yo 8/6/2008 10:05'!
initialize
	"Initialize the receiver"
	| extent |
	pendingUpdate := false.
	zoomOut := false.
	extent := self class virtualScreenExtent.
	Preferences enablePortraitMode ifTrue: [extent := extent transposed].
	super setExtent: extent depth: self class preferredScreenDepth.
	display := DisplayScreen extent: 1@1
			depth: DisplayScreen actualScreenDepth.
	self checkForNewScreenSize.
	self beDisplay; forceToScreen.
	display forceToScreen. "to capture the borders"! !

!OLPCVirtualScreen methodsFor: 'testing' stamp: 'ar 4/13/2007 09:54'!
isVirtualScreen
	"Am I a virtual display screen?"
	^display notNil! !

!OLPCVirtualScreen methodsFor: 'initialize' stamp: 'yo 5/9/2007 10:43'!
prepareToEnter: aProject
	"We need to instantiate the 'correct' OLPC display size here"
	| extent |
	extent := aProject displayExtent.
	extent = self extent ifFalse:[
		super setExtent: extent depth: depth.
		display deferUpdates: true.
		display fillColor: (Color gray: 0.2).
		display deferUpdates: false.
		"ARGH!! This is ugly. We need at least one forceToScreen for display to make sure the pending gray regions get flushed to the display. However, if we do it here, then the project zoom will show flicker (this is why we defer updates in the abov code). But if we don't use the zoom, then we never get any information that the project was successfully entered (and some remnants of the previous project will remain). Therefore, we set a flag here and if the project zoom is triggered then it gets reset. Otherwise, we'll flush the entire display the next time anything gets to be redrawn (which is immediately after entering a new project)"
		pendingUpdate := true.
		self setupWarp.
	].! !

!OLPCVirtualScreen methodsFor: 'copying' stamp: 'bf 3/29/2011 17:05'!
reverse: rect
	super reverse: rect.
	self forceToScreen: rect.! !

!OLPCVirtualScreen methodsFor: 'private' stamp: 'ar 4/13/2007 09:56'!
setExtent: aPoint depth: bitsPerPixel
	"We trap this (otherwise ignored) entry point to get informaed about system startup and display installation."
	display ifNil:[self initialize].! !

!OLPCVirtualScreen methodsFor: 'initialize' stamp: 'ar 4/19/2007 21:09'!
setupWarp
	warpBlt := WarpBlt toForm: display.
	warpBlt sourceForm: self.
	warpBlt cellSize: 2.
	warpBlt combinationRule: Form over.
	warpBlt sourceForm: self destRect: (self displayOffset extent: self extent // self displayScale).
! !

!OLPCVirtualScreen methodsFor: 'initialize' stamp: 'bf 9/21/2007 15:27'!
shutDown
	"Shut me down for saving the image"
	display ifNotNil: [display shutDown].
	bits  := nil.
! !

!OLPCVirtualScreen methodsFor: 'transitions' stamp: 'yo 10/24/2011 11:59'!
zoomIn: goingIn orOutTo: otherImage at: topLeft vanishingPoint: vp
	"Make sure we cover the full display here"
	| otherDisplay otherWarp otherOffset otherScale |
	pendingUpdate := false.
	otherDisplay := Form extent: display extent depth: 32.
	otherDisplay fillColor: (Color gray: 0.2).
	otherScale := (otherImage width / otherDisplay width 
					max: otherImage height / otherDisplay height) max: 1.

	otherOffset := otherDisplay extent - (otherImage extent / otherScale) // 2.

	otherWarp := WarpBlt toForm: otherDisplay.
	otherWarp sourceForm: otherImage.
	otherWarp cellSize: 2.
	otherWarp combinationRule: Form over.
	otherWarp sourceForm: otherImage destRect: (otherOffset extent: otherImage extent // otherScale).
	otherWarp warpBits.
	display zoomIn: goingIn orOutTo: otherDisplay 
		at: 0@0 vanishingPoint: (vp // self displayScale)+self displayOffset.
! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'yo 5/9/2007 12:13'!
zoomOut

	^ zoomOut! !

!OLPCVirtualScreen methodsFor: 'display' stamp: 'yo 5/9/2007 11:57'!
zoomOut: aBoolean

	"When the physical display is bigger than the virtual display size, we have two options.  One is to zoom in and maximize the visible area and another is to map a pixel to a pixel and show it in smaller area (at the center of screen).  This flag governs them."

	self canZoomOut ifFalse: [^ self].
	zoomOut _ aBoolean.
	display fillColor: (Color gray: 0.2). 
	self setupWarp; forceToScreen.
	display forceToScreen. "to capture the borders"
	World restoreMorphicDisplay.
	World repositionFlapsAfterScreenSizeChange.
! !

EventSensor subclass: #OLPCVirtualSensor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OLPC-Display'!

!OLPCVirtualSensor commentStamp: '<historical>' prior: 0!
An input sensor which scales just like the display.!

!OLPCVirtualSensor methodsFor: 'accessing' stamp: 'bf 6/24/2012 14:40'!
primGetNextEvent: array
	| scale offset |
	super primGetNextEvent: array.
	Display isVirtualScreen ifFalse: [^self].
	"scale mouse events"
	(array at: 1) = EventTypeMouse ifTrue:[
		offset := Display displayOffset.
		scale := Display displayScale.
		array at: 3 put: ((array at: 3) - offset x * scale) truncated.
		array at: 4 put: ((array at: 4) - offset y * scale) truncated.
	].
	(array at: 1) = EventTypeDragDropFiles ifTrue:[
		scale := Display displayScale.
		offset := Display displayOffset.
		array at: 4 put: ((array at: 4) - offset x * scale) truncated.
		array at: 5 put: ((array at: 5) - offset y * scale) truncated.
	].! !

!Form methodsFor: '*OLPC-Display' stamp: 'ar 4/12/2007 19:22'!
isVirtualScreen
	"Am I a virtual display screen?"
	^false! !
OLPCVirtualScreen initialize!
